/*! \file DynParams.cpp

    \brief DynRPG parameter-overriding plugin

    \author Bernard J. Badger, AKA Aubrey the Bard

    DynParams is a plugin for the DynRPG project. The purpose of DynParams is
    to allow RPG Maker 2003 users to overwrite the parameters of event script
    commands in RPG Maker 2003. This allows variable values to be used for
    parameters in commands that ordinarily have their parameters set in stone,
    which in turn allows scripting to be more flexible, powerful, and
    data-driven.
*/

#include <DynRPG/DynRPG.h>
#include <assert.h>
#include <iostream>
#include <fstream>


//! Data for a parameter override
/*!
    The ParamPair structure encapsulates the pair of data items needed for a
    parameter override: the index number of the parameter to be overridden, and
    the value to write
*/
struct ParamPair
{
    long lIndex;            //!< Parameter index number
    long lValue;            //!< Value to override parameter with
};

//! Linked list of parameter overrides
/*!
    The ParamList class implements a linked list of parameters to override and
    values to set them to.
*/
class ParamList
{
public:

    //! Default constructor
    /*!
        The default constructor of ParamList provides an empty list.
    */
    ParamList()
    {

        mFront = NULL;
        mBack = NULL;

    }

    //! Destructor
    /*!
        The destructor of ParamList deletes all nodes in the list and the
        ParamList itself.
    */
    ~ParamList()
    {

        ParamListNode * next;     // Temporary pointer to next node

        // Delete all nodes in list
        while( NULL != mFront )
        {

            next = mFront->Next();
            delete mFront;
            mFront = next;

        }

    }

    //! Add a parameter pair at the front of the list
    /*!
        PushFront() adds a parameter pair to the front of the list.

        \param rPair : () The parameter pair being inserted
    */
    void PushFront( ParamPair rPair )
    {

        ParamListNode * node = new ParamListNode(rPair);

        if( Empty() )
        {

            mBack = node;

        }
        mFront = node;

    }

    //! Pop the front parameter pair of the list
    /*!
        PopFront() returns the parameter pair held by the front node of the
        list and deletes that node. It is assumed that a node does exist.

        \return (ParamPair) Front parameter pair of the list
    */
    ParamPair PopFront()
    {

        assert( NULL != mFront );

        ParamPair frontPair;            // Front parameter pair
        ParamListNode * popNode;        // Temporary pointer to node being popped

        // Get parameter pair
        frontPair = mFront->Pair();
        // Delete front node
        popNode = mFront;
        mFront = mFront->Next();
        delete popNode;
        // Return parameter pair
        return frontPair;

    }

    //! Add a parameter pair at the back of the list
    /*!
        PushBack() adds a parameter pair to the back of the list.

        \param rPair : (ParamPair) The parameter pair being inserted
    */
    void PushBack( ParamPair rPair )
    {

        ParamListNode * node = new ParamListNode(rPair);
        if( Empty() )
        {

            mFront = node;
            mBack = node;

        }
        else
        {

            mBack->SetNext(node);
            mBack = node;

        }

    }

    //! Return true if the list is empty
    /*!
        Empty() returns true if the ParamList is an empty list, false otherwise.

        \return (bool) Whether the list is empty
    */
    bool Empty() const
    {

        return mFront == NULL;

    }

    //! Return the parameter pair at the front of the list
    /*!
        Front() returns the parameter pair at the front of the list.  It is
        assumed that the list is not empty.

        \return (ParamPair) The parameter pair at the front of the list
    */
    ParamPair Front()
    {

        assert( NULL != mFront );
        return mFront->Pair();

    }

    //! Return the parameter pair at the back of the list
    /*!
        Back() returns the parameter pair at the back of the list.  It is
        assumed that the list is not empty.

        \return (ParamPair) The parameter pair at the back of the list
    */
    ParamPair Back()
    {

        assert( NULL != mBack );
        return mBack->Pair();

    }

private:

    //! List node
    /*!
        The ParamListNode class is a helper class which implements nodes in the
        linked list. Each node has a parameter index, an override value, and a
        link to the next node.
    */
    class ParamListNode
    {
    public:

        // Default constructor
        ParamListNode()
        {

            mPair.lIndex = 0;
            mPair.lValue = 0;
            mNext = NULL;

        }

        //! Constructor
        /*!
            This constructor creates a ParamListNode with a given parameter
            pair, with no node next in the list.

            \param rPair (ParamPair) Parameter data pair
        */
        ParamListNode( ParamPair rPair )
        {

            mPair = rPair;
            mNext = NULL;

        }

        // Destructor
        ~ParamListNode()
        {

            // Nothing to do

        }

        //! Return mPair
        /*!
            Pair() returns the parameter data pair.

            \return (ParamPair) Parameter index number
        */
        ParamPair Pair()
        {

            return mPair;

        }

        //! Return mNext
        /*!
            Next() returns a pointer to the next node in the list.

            \return (ParamListNode *) Pointer to the next node in the list
        */
        ParamListNode * Next()
        {

            return mNext;

        }

        //! Set the value of mNext
        /*!
            SetNext() sets the pointer to the next node in the list.

            \param next (ParamListNode *) Pointer to the node which should be next in the list
        */
        void SetNext( ParamListNode * rNext )
        {

            mNext = rNext;

        }

    private:

        ParamPair mPair;            //!< Parameter data pair
        ParamListNode * mNext;      //!< Pointer to next node in list

    };

    //! Return mFront
    /*!
        FrontNode() returns a pointer to the front node of the list.

        \return (ParamListNode *) Pointer to front node of list
    */
    ParamListNode * FrontNode() const
    {

        return mFront;

    }

    //! Return mBack
    /*!
        BackNode() returns a pointer to the back node of the list.

        \return (ParamListNode *) Pointer to back node of list
    */
    ParamListNode * BackNode() const
    {

        return mBack;

    }

    ParamListNode * mFront;         //!< Pointer to front node of the list
    ParamListNode * mBack;          //!< Pointer to back node of the list

};

// Plugin global variables

ParamList paramList;            //!< List of parameter data pairs for next event command override
std::string stringParam;        //!< String parameter to use in next event command override
long lBuildIndex;               //!< Index for adding parameters to the list with ambiguous positioning
const long clDefaultBuildIndex = 5;     //!< Default build index (5 is the starting index of subcommands in Move Event event commands)

bool bRecordParams;             //!< Flag to indicate whether to record parameters to the output file
bool bOverwriteParams;          //!< Flag to indicate whether the next event command should have its parameters overwritten
bool bOverwriteString;          //!< Flag to indicate whether the string parameter should be overwritten (only takes effect if bOverwriteParams is also true)

std::ofstream ofParams;         //!< File output stream to record parameters



//! Clear the accumulated overwrite parameters
/*!
    clearParams() is a function which deletes all of the overwrite parameters,
    both number parameters and the string parameter. It is called after the
    plugin is used to overwrite the parameters of an event command, and it can
    also be called by a comment command.
*/
void clearParams()
{

    stringParam.clear();
    bOverwriteString = false;
    bOverwriteParams = false;
    lBuildIndex = clDefaultBuildIndex;
    while( !paramList.Empty() )
    {   // Pop number parameters until list is empty

        paramList.PopFront();

    }

}

//! Record the parameters of an event command
/*!
    recordParameters() writes the parameters of a DynRPG EventScriptLine to a
    text file. This is useful for analyzing the workings of a given command or
    quickly determining what the parameters are like for complicated cases,
    such as the ASCII code integer parameters used in a Move Event command when
    referencing a file name.

    \param scriptLine (RPG::EventScriptLine*) Pointer to the scriptLine for the event command being recorded
*/
static void recordParameters( RPG::EventScriptLine *scriptLine )
{

    static int i;       // Increment variable

    if( !ofParams.is_open() )
    {   // Output file stream has not been initialized

        ofParams.open( "DynPlugins\\DynParamsRecord.txt" );

    }
    ofParams << "EventCommand type: " << scriptLine->command << "\n";
    ofParams << "String parameter: \"" << scriptLine->stringParameter.s_str() << "\"\n";
    ofParams << "Number parameters:\n";
    for( i = 0; i < scriptLine->parameters.size; i++ )
    {

        ofParams << i + 1 << ": " << scriptLine->parameters.items[i] << "\n";

    }
    ofParams << "\n";

}

//! Initialize the plugin
/*!
    onInitTitleScreen() is called when the game goes to the title screen. In
    this plugin this is used to initialize things.
*/
void onInitTitleScreen()
{

    // Initialize variables
    stringParam.clear();
    bRecordParams = false;
    bOverwriteParams = false;
    bOverwriteString = false;
    lBuildIndex = clDefaultBuildIndex;

}

//! Clean up after use
/*!
    onExit() is called when the game closes. In this plugin this is used to
    perform any needed cleanup.
*/
void onExit()
{

    // Close the parameter scan output file
    ofParams.close();

}

//! Respond to potential comment commands
/*!
    onComment() is called when the game runs across a comment line in its event
    scripting. Comment commands are how the user communicates with the plugin,
    so this function checks for comment commands and responds accordingly.

    \param text (const char*) The comment's content as simple text
    \param parsedData (const RPG::ParsedCommentData*) The already parsed data
    \param nextScriptLine (RPG::EventScriptLine*) The next event script line after the comment
    \param scriptData (RPG::EventScriptData*) Pointer to the RPG::EventScriptData object of the current event script
    \param eventId (int) The ID of the current event (negative for common events, zero for battle events)
    \param pageId (int) The ID of the current event page (zero for common and battle events - sorry, no battle event page ID yet)
    \param lineId (int) The zero-based line number
    \param nextLineId (int*) Pointer to the next executed line number (-1 for default)
    \return (bool) false will prevent other plugins from receiving this notification, use true otherwise
*/
bool onComment( const char* text,
                const RPG::ParsedCommentData* 	parsedData,
                RPG::EventScriptLine* 	nextScriptLine,
                RPG::EventScriptData* 	scriptData,
                int 	eventId,
                int 	pageId,
                int 	lineId,
                int* 	nextLineId )
{

    std::string cmd = parsedData->command;
    std::string commentText = text;
    static ParamPair newPair;
    char buffer [33];

    // Check for and respond to plugin comment commands
    if( 0 == cmd.compare( "dynparams_append_string" ) )
    {   // Append text to the string parameter

        //ofParams << "Text parameter: \"" << commentText << "\"\n";      // DEBUG
        stringParam += commentText.substr( 25 );        // 25 is the index of the first character after "@dynparams_append_string " in the comment command
        //ofParams << "New string value: \"" << stringParam << "\"\n";      // DEBUG
        bOverwriteString = true;

    }
    if( 0 == cmd.compare( "dynparams_append_number" ) )
    {   // Append a number to the string parameter

        stringParam += itoa( parsedData->parameters[0].number, buffer, 10 );
        bOverwriteString = true;

    }
    if( 0 == cmd.compare( "dynparams_add_param" ) )
    {   // Add a parameter data pair to the list

        if( 1 == parsedData->parametersCount )
        {   // Only one comment command parameter; add to overwrite parameter list using list builder index

            newPair.lIndex = lBuildIndex;
            newPair.lValue = parsedData->parameters[0].number;
            lBuildIndex++;

        }
        else
        {   // More than one comment command parameter; add to overwrite parameter list using given index

            newPair.lIndex = parsedData->parameters[0].number;
            newPair.lValue = parsedData->parameters[1].number;

        }
        //ofParams << "Pushing parameter pair: " << newPair.lIndex << ", " << newPair.lValue << "\n";      // DEBUG
        paramList.PushBack( newPair );
        return false;

    }
    if( 0 == cmd.compare( "dynparams_set_index" ) )
    {   // Set the starting index for adding successive parameters to the list

        lBuildIndex = parsedData->parameters[0].number;
        return false;

    }
    if( 0 == cmd.compare( "dynparams_clear_params" ) )
    {   // Clear the list of parameter data pairs and the string parameter

        clearParams();
        return false;

    }
    if( 0 == cmd.compare( "dynparams_start_record" ) )
    {   // Start recording the parameters of event commands to the output file

        bRecordParams = true;
        return false;

    }
    if( 0 == cmd.compare( "dynparams_stop_record" ) )
    {   // Stop recording the parameters of event commands to the output file

        bRecordParams = false;
        return false;

    }
    if( 0 == cmd.compare( "dynparams_overwrite_next" ) )
    {   // Overwrite the parameters of the next event command according to plugin's parameter list

        bOverwriteParams = true;
        return false;

    }
    // No comment commands for this plugin detected; pass notification on for other plugins
    return true;

}

//! Respond to potential command overwrite cases
/*!
    onEventCommand() is called when the game runs an event command line. The
    essence of the DynParams plugin is to overwrite the parameters of a given
    command. This function checks whether the current command is of a type that
    can be overridden (anything besides an empty line, comment, or "end" line)
    and whether the user has indicated that this command should be overridden.
    If so, it overwrites the parameters of the command with the data stored
    due to preceding comment commands.

    \param scriptLine (RPG::EventScriptLine*) The event script line which is about to be executed
    \param scriptData (RPG::EventScriptData*) Pointer to the RPG::EventScriptData object of the current event script
    \param eventId (int) The ID of the current event (negative for common events, zero for battle events)
    \param pageId (int) The ID of the current event page (zero for common and battle events - sorry, no battle event page ID yet)
    \param lineId (int) The zero-based line number
    \param nextLineId (int*) Pointer to the next executed line number (-1 for default)
    \return (bool) false will prevent other plugins from receiving this notification, use true otherwise
*/
bool onEventCommand ( RPG::EventScriptLine *scriptLine,
                     RPG::EventScriptData *scriptData,
                     int eventId, int pageId,
                     int lineId, int *nextLineId )
{

    static ParamPair curPair;           // Parameter pair for applying parameter overwrites

    if( bOverwriteParams )
    {   // Overwrite parameters flag is on; check whether the event command is of a valid type for overwriting

        if( scriptLine->command != RPG::EVCMD_END_OF_EVENT &&
            scriptLine->command != RPG::EVCMD_EMPTY_LINE &&
            scriptLine->command != RPG::EVCMD_CHOICE_END &&
            scriptLine->command != RPG::EVCMD_BATTLE_END &&
            scriptLine->command != RPG::EVCMD_SHOP_END &&
            scriptLine->command != RPG::EVCMD_INN_END &&
            scriptLine->command != RPG::EVCMD_FORK_END &&
            scriptLine->command != RPG::EVCMD_END_LOOP &&
            scriptLine->command != RPG::EVCMD_FORK_IN_BATTLE_END )
        {   // Event command is valid for overwriting, proceed with overwrite

            if( bOverwriteString )
            {   // Overwrite string flag is on; overwrite string parameter

                scriptLine->stringParameter = stringParam;

            }
            while( !paramList.Empty() )
            {   // Overwrite number parameters until list is empty

                curPair = paramList.PopFront();
                //ofParams << "Popped parameter pair: " << curPair.lIndex << ", " << curPair.lValue << "\n";      // DEBUG
                scriptLine->parameters.items[curPair.lIndex - 1] = curPair.lValue;

            }
            // Clear parameters and reset overwrite parameters flag
            clearParams();
            bOverwriteParams = false;

        }

    }
    if( bRecordParams )
    {   // Record parameters flag is on; call function to record parameters

        recordParameters( scriptLine );

    }
    return true;

}
